#include <windows.h>
#include <stdio.h>

#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"

#include "resource.h"
#include "filter.h"

///////////////////////////////////////////////////////////////////////////

int tutorialRunProc(const FilterActivation *fa, const FilterFunctions *ff);
int tutorialStartProc(FilterActivation *fa, const FilterFunctions *ff);
int tutorialEndProc(FilterActivation *fa, const FilterFunctions *ff);
long tutorialParamProc(FilterActivation *fa, const FilterFunctions *ff);
int tutorialConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
void tutorialStringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str);
void tutorialScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool tutorialFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);

///////////////////////////////////////////////////////////////////////////

typedef struct MyFilterData {
    Pixel32 *grn_tab;
    Pixel32 *blu_tab;
    bool     fExpand;
    bool     fThird;
} MyFilterData;

ScriptFunctionDef tutorial_func_defs[]={
    { (ScriptFunctionPtr)tutorialScriptConfig, "Config", "0ii" },
    { NULL },
};

CScriptObject tutorial_obj={
    NULL, tutorial_func_defs
};

struct FilterDefinition filterDef_tutorial = {

    NULL, NULL, NULL,       // next, prev, module
    "tutorial",             // name
    "Emphasizes green and de-emphasizes blue in the image.",
                            // desc
    "anyone",               // maker
    NULL,                   // private_data
    sizeof(MyFilterData),   // inst_data_size

    NULL,                   // initProc
    NULL,                   // deinitProc
    tutorialRunProc,        // runProc
    tutorialParamProc,      // paramProc
    tutorialConfigProc,     // configProc
    tutorialStringProc,     // stringProc
    tutorialStartProc,      // startProc
    tutorialEndProc,        // endProc

    &tutorial_obj,          // script_obj
    tutorialFssProc,        // fssProc

};

///////////////////////////////////////////////////////////////////////////

extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

static FilterDefinition *fd_tutorial;

///////////////////////////////////////////////////////////////////////////

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
    if (!(fd_tutorial = ff->addFilter(fm, &filterDef_tutorial, sizeof(FilterDefinition))))
        return 1;

	vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
	vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

    return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
    ff->removeFilter(fd_tutorial);
}

///////////////////////////////////////////////////////////////////////////

static bool g_MMXenabled;

int tutorialStartProc(FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
    int i;

    g_MMXenabled = ff->isMMXEnabled();

    if (!(mfd->grn_tab = new Pixel32[256 * 2]))
        return 1;

    mfd->blu_tab = mfd->grn_tab + 256;

    if (mfd->fThird)
        for(i=0; i<256; i++) {
            mfd->grn_tab[i] = ((0x200 + i) / 3) << 8;
            mfd->blu_tab[i] = i / 3;
        }
    else
        for(i=0; i<256; i++) {
            mfd->grn_tab[i] = ((0x100 + i) / 2) << 8;
            mfd->blu_tab[i] = i / 2;
        }

    return 0;
}

void __declspec(naked) doscan_MMX(Pixel32 *dst, Pixel32 *src, int w, long frac, long bias, int fDouble) {
    static const __int64 Rmask = 0x0000FFFF00000000i64;

    __asm {
        push        ebp
        push        edi
        push        esi
        push        ebx

        mov         eax,[esp+4+16]
        mov         edx,[esp+8+16]

        mov         ecx,[esp+12+16]
        neg         ecx
        shl         ecx,2
        sub         eax,ecx
        sub         edx,ecx

        movq        mm6,Rmask
        movd        mm4,[esp+20+16]
        psllq       mm4,16
        movd        mm5,[esp+16+16]
        punpcklwd   mm5,mm5
        pxor        mm7,mm7

        mov         ebx,dword ptr [esp+24+16]
        or          ebx,ebx
        jz          xloop1

        sub         eax,ecx

xloop2:
        movd        mm0,[edx+ecx]       ;mm0 = pixel
        movq        mm1,mm6             ;mm1 = R mask
        punpcklbw   mm0,mm7             ;unpack pixel to words
        pand        mm1,mm0             ;mm1 = red component
        pmulhw      mm0,mm5             ;scale green and blue
        paddw       mm1,mm4             ;add green bias
        paddw       mm0,mm1             ;add scaled green/blue
        packuswb    mm0,mm0             ;repack pixel to bytes
        movq        [eax+ecx*2],mm0     ;write 2 pixels
        add         ecx,4
        jne         xloop2
		jmp			short xit

xloop1:
        movd        mm0,[edx+ecx]       ;mm0 = pixel
        movq        mm1,mm6             ;mm1 = R mask
        punpcklbw   mm0,mm7             ;unpack pixel to words
        pand        mm1,mm0             ;mm1 = red component
        paddw       mm0,mm0             ;double g/b channels beforehand
        pmulhw      mm0,mm5             ;scale green and blue
        paddw       mm1,mm4             ;add green bias
        paddw       mm0,mm1             ;add scaled green/blue
        packuswb    mm0,mm0             ;repack pixel to bytes
        movd        [eax+ecx],mm0       ;write pixel
        add         ecx,4
        jne         xloop1

xit:
        pop         ebx
        pop         esi
        pop         edi
        pop         ebp
        ret
    }
}

int tutorialRunProc(const FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
    PixDim w, h;
    Pixel32 *src, *dst;
    const Pixel32 *grn_tab = mfd->grn_tab;
    const Pixel32 *blu_tab = mfd->blu_tab;

    src = (Pixel32 *)fa->src.data;
    dst = (Pixel32 *)fa->dst.data;

    h = fa->src.h;
    do {
        w = fa->src.w;

        if (g_MMXenabled) {

            doscan_MMX(dst, src, w,
                    mfd->fThird ? 0x2AAA : 0x4000,
                    mfd->fThird ? 0x55 : 0x80,
                    mfd->fExpand);


		    src = (Pixel32 *)((char *)src + fa->src.pitch);
	        dst = (Pixel32 *)((char *)dst + fa->dst.pitch);

    // double the routine for speed; an if would kill us in the
    // inner loop, but in the outer loop it's ok

        } else {
			if (mfd->fExpand)
				do {
					Pixel32 old_pixel, new_pixel;

					old_pixel = *src++;

					new_pixel = (old_pixel & 0xFF0000)
							  + grn_tab[(old_pixel>>8) & 0xff]
							  + blu_tab[old_pixel & 0xff];

					*dst++ = new_pixel;
					*dst++ = new_pixel;
				} while(--w);
			else
				do {
					Pixel32 old_pixel, new_pixel;

					old_pixel = *src++;

					new_pixel = (old_pixel & 0xFF0000)
							  + grn_tab[(old_pixel>>8) & 0xff]
							  + blu_tab[old_pixel & 0xff];

					*dst++ = new_pixel;
				} while(--w);

	        src = (Pixel32 *)((char *)src + fa->src.modulo);
		    dst = (Pixel32 *)((char *)dst + fa->dst.modulo);
		}

    } while(--h);

    if (g_MMXenabled)
		__asm emms

    return 0;
}


int tutorialEndProc(FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    delete[] mfd->grn_tab;
	mfd->grn_tab = NULL;

    return 0;
}

long tutorialParamProc(FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    if (mfd->fExpand)
	    fa->dst.w *= 2;

	fa->dst.pitch = (fa->dst.w*4 + 7) & -8;

    return FILTERPARAM_SWAP_BUFFERS;
}


BOOL CALLBACK tutorialConfigDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) {
    MyFilterData *mfd = (MyFilterData *)GetWindowLong(hdlg, DWL_USER);

    switch(msg) {
        case WM_INITDIALOG:
            SetWindowLong(hdlg, DWL_USER, lParam);
            mfd = (MyFilterData *)lParam;

            CheckDlgButton(hdlg, IDC_EXPAND, mfd->fExpand?BST_CHECKED:BST_UNCHECKED);
            CheckDlgButton(hdlg, IDC_THIRD, mfd->fThird?BST_CHECKED:BST_UNCHECKED);

            return TRUE;

        case WM_COMMAND:
            switch(LOWORD(wParam)) {
            case IDOK:
                mfd->fExpand = !!IsDlgButtonChecked(hdlg, IDC_EXPAND);
                mfd->fThird = !!IsDlgButtonChecked(hdlg, IDC_THIRD);
                EndDialog(hdlg, 0);
                return TRUE;
            case IDCANCEL:
                EndDialog(hdlg, 1);
                return FALSE;
            }
            break;
    }

    return FALSE;
}

int tutorialConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd) {
    return DialogBoxParam(fa->filter->module->hInstModule,
            MAKEINTRESOURCE(IDD_FILTER_TUTORIAL), hwnd,
            tutorialConfigDlgProc, (LPARAM)fa->filter_data);
}

void tutorialStringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str) {
    const char *modes[2][2]={
        " (normal)",
        " (expand)",
        " (third)",
        " (expand+third)",
    };
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    strcpy(str, modes[mfd->fThird][mfd->fExpand]);
}

void tutorialScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {
    FilterActivation *fa = (FilterActivation *)lpVoid;
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    mfd->fExpand    = !!argv[0].asInt();
    mfd->fThird     = !!argv[1].asInt();
}

bool tutorialFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    _snprintf(buf, buflen, "Config(%d, %d)",
        mfd->fExpand,
        mfd->fThird);

    return true;
}

